from dataclasses import dataclass, field, replace
from datetime import datetime
from distutils import util
from enum import Enum
from typing import Any, List, Optional, Protocol, runtime_checkable, Sequence

from .chaum_pedersen import (
    ConstantChaumPedersenProof,
    DisjunctiveChaumPedersenProof,
    make_constant_chaum_pedersen,
    make_disjunctive_chaum_pedersen,
)
from .election_object_base import ElectionObjectBase
from .elgamal import ElGamalCiphertext, elgamal_add
from .group import add_q, ElementModP, ElementModQ, ZERO_MOD_Q
from .hash import CryptoHashCheckable, hash_elems
from .logs import log_warning
from .tracker import get_rotating_tracker_hash, tracker_hash_to_words
from .utils import to_ticks, flatmap_optional


def _list_eq(
    list1: Sequence[ElectionObjectBase], list2: Sequence[ElectionObjectBase]
) -> bool:
    """
    We want to compare lists of election objects as if they're sets. We fake this by first
    sorting them on their object ids, then using regular list comparison.
    """
    return sorted(list1, key=lambda x: x.object_id) == sorted(
        list2, key=lambda x: x.object_id
    )


@dataclass(eq=True, unsafe_hash=True)
class ExtendedData(object):
    """
    ExtendedData represents any arbitrary data expressible as a string with a length.

    This class is used primarily as a field on a selection to indicate a write-in candidate text value
    """

    value: str
    length: int


@dataclass(unsafe_hash=True)
class PlaintextBallotSelection(ElectionObjectBase):
    """
    A BallotSelection represents an individual selection on a ballot.

    This class accepts a `vote` string field which has no constraints
    in the ElectionGuard Data Specification, but is constrained logically
    in the application to resolve to `True` or `False`.  This implies that the
    data specification supports passing any string that can be represented as
    an integer, however only 0 and 1 is supported for now.

    This class can also be designated as `is_placeholder_selection` which has no
    context to the data specification but is useful for running validity checks internally

    an `extended_data` field exists to support any arbitrary data to be associated
    with the selection.  In practice, this field is the cleartext representation
    of a write-in candidate value.  In the current implementation these values are
    discarded when encrypting.
    """

    vote: str

    is_placeholder_selection: bool = field(default=False)
    """Determines if this is a placeholder selection"""

    # TODO: ISSUE #35: encrypt/decrypt
    extended_data: Optional[ExtendedData] = field(default=None)
    """
    an optional field of arbitrary data, such as the value of a write-in candidate
    """

    def is_valid(self, expected_object_id: str) -> bool:
        """
        Given a PlaintextBallotSelection validates that the object matches an expected object
        and that the plaintext string can resolve to a valid representation
        """

        if self.object_id != expected_object_id:
            log_warning(
                f"invalid object_id: expected({expected_object_id}) actual({self.object_id})"
            )
            return False

        choice = self.to_int()
        if choice < 0 or choice > 1:
            log_warning(f"Currently only supporting choices of 0 or 1: {str(self)}")
            return False

        return True

    def to_int(self) -> int:
        """
        Represent a Truthy string as 1, or if the string is Falsy, 0
        See: https://docs.python.org/3/distutils/apiref.html#distutils.util.strtobool

        :return: an integer 0 or 1 for valid data, or 0 if the data is malformed
        """

        as_bool = False
        try:
            as_bool = util.strtobool(self.vote.lower())
        except ValueError:
            log_warning(
                f"to_int could not convert plaintext: {self.vote.lower()} to bool"
            )

        # TODO: ISSUE #33: If the boolean coercion above fails, support integer votes
        # greater than 1 for cases such as cumulative voting
        as_int = int(as_bool)
        return as_int

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, PlaintextBallotSelection)
            and self.object_id == other.object_id
            and self.vote == other.vote
            and self.is_placeholder_selection == other.is_placeholder_selection
            and self.extended_data == other.extended_data
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


@runtime_checkable
class CiphertextSelection(Protocol):
    """
    Encrypted selection
    """

    object_id: str

    description_hash: ElementModQ
    """The SelectionDescription hash"""

    ciphertext: ElGamalCiphertext
    """The encrypted representation of the selection"""


@dataclass(eq=True, unsafe_hash=True)
class CiphertextBallotSelection(
    ElectionObjectBase, CiphertextSelection, CryptoHashCheckable
):
    """
    A CiphertextBallotSelection represents an individual encrypted selection on a ballot.

    This class accepts a `description_hash` and a `ciphertext` as required parameters
    in its constructor.

    When a selection is encrypted, the `description_hash` and `ciphertext` required fields must
    be populated at construction however the `nonce` is also usually provided by convention.

    After construction, the `crypto_hash` field is populated automatically in the `__post_init__` cycle

    A consumer of this object has the option to discard the `nonce` and/or discard the `proof`,
    or keep both values.  
    
    By discarding the `nonce`, the encrypted representation and `proof`
    can only be regenerated if the nonce was derived from the ballot's master nonce.  If the nonce
    used for this selection is truly random, and it is discarded, then the proofs cannot be regenerated.

    By keeping the `nonce`, or deriving the selection nonce from the ballot nonce, an external system can 
    regenerate the proofs on demand.  This is useful for storage or memory constrained systems.

    By keeping the `proof` the nonce is not required fotor verify the encrypted selection.
    """

    description_hash: ElementModQ
    """The SelectionDescription hash"""

    ciphertext: ElGamalCiphertext
    """The encrypted representation of the vote field"""

    crypto_hash: ElementModQ
    """The hash of the encrypted values"""

    is_placeholder_selection: bool = field(default=False)
    """Determines if this is a placeholder selection"""

    nonce: Optional[ElementModQ] = field(default=None)
    """The nonce used to generate the encryption. Sensitive & should be treated as a secret"""

    proof: Optional[DisjunctiveChaumPedersenProof] = field(default=None)
    """The proof that demonstrates the selection is an encryption of 0 or 1, and was encrypted using the `nonce`"""

    # TODO: ISSUE #35: encrypt/decrypt
    extended_data: Optional[ElGamalCiphertext] = field(default=None)
    """encrypted representation of the extended_data field"""

    def is_valid_encryption(
        self,
        seed_hash: ElementModQ,
        elgamal_public_key: ElementModP,
        crypto_extended_base_hash: ElementModQ,
    ) -> bool:
        """
        Given an encrypted BallotSelection, validates the encryption state against a specific seed hash and public key.
        Calling this function expects that the object is in a well-formed encrypted state
        with the elgamal encrypted `message` field populated along with the DisjunctiveChaumPedersenProof `proof` populated.
        the ElementModQ `description_hash` and the ElementModQ `crypto_hash` are also checked. 

        :param seed_hash: the hash of the SelectionDescription, or 
                          whatever `ElementModQ` was used to populate the `description_hash` field.
        :param elgamal_public_key: The election public key
        """

        if seed_hash != self.description_hash:
            log_warning(
                f"mismatching selection hash: {self.object_id} expected({str(seed_hash)}), actual({str(self.description_hash)})"
            )
            return False

        recalculated_crypto_hash = self.crypto_hash_with(seed_hash)
        if self.crypto_hash != recalculated_crypto_hash:
            log_warning(
                f"mismatching crypto hash: {self.object_id} expected({str(recalculated_crypto_hash)}), actual({str(self.crypto_hash)})"
            )
            return False

        if self.proof is None:
            log_warning(f"no proof exists for: {self.object_id}")
            return False

        return self.proof.is_valid(
            self.ciphertext, elgamal_public_key, crypto_extended_base_hash
        )

    def crypto_hash_with(self, seed_hash: ElementModQ) -> ElementModQ:
        """
        Given an encrypted BallotSelection, generates a hash, suitable for rolling up
        into a hash / tracking code for an entire ballot. Of note, this particular hash examines
        the `seed_hash` and `message`, but not the proof.
        This is deliberate, allowing for the possibility of ElectionGuard variants running on
        much more limited hardware, wherein the Disjunctive Chaum-Pedersen proofs might be computed
        later on.

        In most cases the seed_hash should match the `description_hash`
        """
        return _ciphertext_ballot_selection_crypto_hash_with(
            self.object_id, seed_hash, self.ciphertext
        )


def _ciphertext_ballot_selection_crypto_hash_with(
    object_id: str, seed_hash: ElementModQ, ciphertext: ElGamalCiphertext
) -> ElementModQ:
    return hash_elems(object_id, seed_hash, ciphertext.crypto_hash())


def make_ciphertext_ballot_selection(
    object_id: str,
    description_hash: ElementModQ,
    ciphertext: ElGamalCiphertext,
    elgamal_public_key: ElementModP,
    crypto_extended_base_hash: ElementModQ,
    proof_seed: ElementModQ,
    selection_representation: int,
    is_placeholder_selection: bool = False,
    nonce: Optional[ElementModQ] = None,
    crypto_hash: Optional[ElementModQ] = None,
    proof: Optional[DisjunctiveChaumPedersenProof] = None,
    extended_data: Optional[ElGamalCiphertext] = None,
) -> CiphertextBallotSelection:
    """
    Constructs a `CipherTextBallotSelection` object. Most of the parameters here match up to fields
    in the class, but this helper function will optionally compute a Chaum-Pedersen proof if the
    given nonce isn't `None`. Likewise, if a crypto_hash is not provided, it will be derived from
    the other fields.
    """
    if crypto_hash is None:
        crypto_hash = _ciphertext_ballot_selection_crypto_hash_with(
            object_id, description_hash, ciphertext
        )

    if proof is None:
        proof = flatmap_optional(
            nonce,
            lambda n: make_disjunctive_chaum_pedersen(
                ciphertext,
                n,
                elgamal_public_key,
                crypto_extended_base_hash,
                proof_seed,
                selection_representation,
            ),
        )

    return CiphertextBallotSelection(
        object_id=object_id,
        description_hash=description_hash,
        ciphertext=ciphertext,
        is_placeholder_selection=is_placeholder_selection,
        nonce=nonce,
        crypto_hash=crypto_hash,
        proof=proof,
        extended_data=extended_data,
    )


@dataclass(unsafe_hash=True)
class PlaintextBallotContest(ElectionObjectBase):
    """
    A PlaintextBallotContest represents the selections made by a voter for a specific ContestDescription

    this class can be either a partial or a complete representation of a contest dataset.  Specifically,
    a partial representation must include at a minimum the "affirmative" selections of a contest.
    A complete representation of a ballot must include both affirmative and negative selections of
    the contest, AND the placeholder selections necessary to satisfy the ConstantChaumPedersen proof 
    in the CiphertextBallotContest.

    Typically partial contests are passed into Electionguard for memory constrained systems, 
    while complete contests are passed into ElectionGuard when running encryption on an existing dataset.
    """

    ballot_selections: List[PlaintextBallotSelection] = field(
        default_factory=lambda: []
    )
    """Collection of ballot selections"""

    def is_valid(
        self,
        expected_object_id: str,
        expected_number_selections: int,
        expected_number_elected: int,
        votes_allowed: Optional[int] = None,
    ) -> bool:
        """
        Given a PlaintextBallotContest returns true if the state is representative of the expected values.

        Note: because this class supports partial representations, undervotes are considered a valid state.
        """

        if self.object_id != expected_object_id:
            log_warning(
                f"invalid object_id: expected({expected_object_id}) actual({self.object_id})"
            )
            return False

        if len(self.ballot_selections) > expected_number_selections:
            log_warning(
                f"invalid number_selections: expected({expected_number_selections}) actual({len(self.ballot_selections)})"
            )
            return False

        number_elected = 0
        votes = 0

        # Verify the selections are well-formed
        for selection in self.ballot_selections:
            selection_count = selection.to_int()
            votes += selection_count
            if selection_count >= 1:
                number_elected += 1

        if number_elected > expected_number_elected:
            log_warning(
                f"invalid number_elected: expected({expected_number_elected}) actual({number_elected})"
            )
            return False

        if votes_allowed is not None and votes > votes_allowed:
            log_warning(f"invalid votes: expected({votes_allowed}) actual({votes})")
            return False

        return True

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, PlaintextBallotContest) and _list_eq(
            self.ballot_selections, other.ballot_selections
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


@dataclass(unsafe_hash=True)
class CiphertextBallotContest(ElectionObjectBase, CryptoHashCheckable):
    """
    A CiphertextBallotContest represents the selections made by a voter for a specific ContestDescription

    CiphertextBallotContest can only be a complete representation of a contest dataset.  While
    PlaintextBallotContest supports a partial representation, a CiphertextBallotContest includes all data
    necessary for a verifier to verify the contest.  Specifically, it includes both explicit affirmative 
    and negative selections of the contest, as well as the placeholder selections that satisfy
    the ConstantChaumPedersen proof.

    Similar to `CiphertextBallotSelection` the consuming application can choose to discard or keep both
    the `nonce` and the `proof` in some circumstances.  For deterministic nonce's derived from the 
    master nonce, both values can be regenerated.  If the `nonce` for this contest is completely random,
    then it is required in order to regenerate the proof.
    """

    description_hash: ElementModQ
    """Hash from contestDescription"""

    ballot_selections: List[CiphertextBallotSelection]
    """Collection of ballot selections"""

    crypto_hash: ElementModQ
    """Hash of the encrypted values"""

    nonce: Optional[ElementModQ] = None
    """The nonce used to generate the encryption. Sensitive & should be treated as a secret"""

    proof: Optional[ConstantChaumPedersenProof] = None
    """
    The proof demonstrates the sum of the selections does not exceed the maximum
    available selections for the contest, and that the proof was generated with the nonce
    """

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, CiphertextBallotContest)
            and self.object_id == other.object_id
            and _list_eq(self.ballot_selections, other.ballot_selections)
            and self.description_hash == other.description_hash
            and self.crypto_hash == other.crypto_hash
            and self.nonce == other.nonce
            and self.proof == other.proof
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def aggregate_nonce(self) -> Optional[ElementModQ]:
        """
        :return: an aggregate nonce for the contest composed of the nonces of the selections
        """
        return _ciphertext_ballot_contest_aggregate_nonce(
            self.object_id, self.ballot_selections
        )

    def crypto_hash_with(self, seed_hash: ElementModQ) -> ElementModQ:
        """
        Given an encrypted BallotContest, generates a hash, suitable for rolling up
        into a hash / tracking code for an entire ballot. Of note, this particular hash examines
        the `seed_hash` and `ballot_selections`, but not the proof.
        This is deliberate, allowing for the possibility of ElectionGuard variants running on
        much more limited hardware, wherein the Disjunctive Chaum-Pedersen proofs might be computed
        later on.

        In most cases, the seed_hash is the description_hash
        """
        return _ciphertext_ballot_context_crypto_hash(
            self.object_id, self.ballot_selections, seed_hash
        )

    def elgamal_accumulate(self) -> ElGamalCiphertext:
        """
        Add the individual ballot_selections `message` fields together, suitable for use
        in a Chaum-Pedersen proof.
        """
        return _ciphertext_ballot_elgamal_accumulate(self.ballot_selections)

    def is_valid_encryption(
        self,
        seed_hash: ElementModQ,
        elgamal_public_key: ElementModP,
        crypto_extended_base_hash: ElementModQ,
    ) -> bool:
        """
        Given an encrypted BallotContest, validates the encryption state against a specific seed hash and public key
        by verifying the accumulated sum of selections match the proof.
        Calling this function expects that the object is in a well-formed encrypted state
        with the `ballot_selections` populated with valid encrypted ballot selections,
        the ElementModQ `description_hash`, the ElementModQ `crypto_hash`, and the ConstantChaumPedersenProof all populated. 
        Specifically, the seed hash in this context is the hash of the ContestDescription, 
        or whatever `ElementModQ` was used to populate the `description_hash` field.
        """
        if seed_hash != self.description_hash:
            log_warning(
                f"mismatching contest hash: {self.object_id} expected({str(seed_hash)}), actual({str(self.description_hash)})"
            )
            return False

        recalculated_crypto_hash = self.crypto_hash_with(seed_hash)
        if self.crypto_hash != recalculated_crypto_hash:
            log_warning(
                f"mismatching crypto hash: {self.object_id} expected({str(recalculated_crypto_hash)}), actual({str(self.crypto_hash)})"
            )
            return False

        # NOTE: this check does not verify the proofs of the individual selections by design.

        if self.proof is None:
            log_warning(f"no proof exists for: {self.object_id}")
            return False

        # Verify the sum of the selections matches the proof
        elgamal_accumulation = self.elgamal_accumulate()
        return self.proof.is_valid(
            elgamal_accumulation, elgamal_public_key, crypto_extended_base_hash
        )


def _ciphertext_ballot_elgamal_accumulate(
    ballot_selections: List[CiphertextBallotSelection],
) -> ElGamalCiphertext:
    return elgamal_add(*[selection.ciphertext for selection in ballot_selections])


def _ciphertext_ballot_context_crypto_hash(
    object_id: str,
    ballot_selections: List[CiphertextBallotSelection],
    seed_hash: ElementModQ,
) -> ElementModQ:
    if len(ballot_selections) == 0:
        log_warning(
            f"mismatching ballot_selections state: {object_id} expected(some), actual(none)"
        )
        return ZERO_MOD_Q

    selection_hashes = [selection.crypto_hash for selection in ballot_selections]

    return hash_elems(object_id, seed_hash, *selection_hashes)


def _ciphertext_ballot_contest_aggregate_nonce(
    object_id: str, ballot_selections: List[CiphertextBallotSelection]
) -> Optional[ElementModQ]:
    selection_nonces: List[ElementModQ] = list()
    for selection in ballot_selections:
        if selection.nonce is None:
            log_warning(
                f"missing nonce values for contest {object_id} cannot calculate aggregate nonce"
            )
            return None
        else:
            selection_nonces.append(selection.nonce)

    return add_q(*selection_nonces)


def make_ciphertext_ballot_contest(
    object_id: str,
    description_hash: ElementModQ,
    ballot_selections: List[CiphertextBallotSelection],
    elgamal_public_key: ElementModP,
    crypto_extended_base_hash: ElementModQ,
    proof_seed: ElementModQ,
    number_elected: int,
    crypto_hash: Optional[ElementModQ] = None,
    proof: Optional[ConstantChaumPedersenProof] = None,
    nonce: Optional[ElementModQ] = None,
) -> CiphertextBallotContest:
    """
    Constructs a `CipherTextBallotContest` object. Most of the parameters here match up to fields
    in the class, but this helper function will optionally compute a Chaum-Pedersen proof if the
    ballot selections include their encryption nonces. Likewise, if a crypto_hash is not provided,
    it will be derived from the other fields.
    """
    if crypto_hash is None:
        crypto_hash = _ciphertext_ballot_context_crypto_hash(
            object_id, ballot_selections, description_hash
        )

    aggregate = _ciphertext_ballot_contest_aggregate_nonce(object_id, ballot_selections)
    if proof is None:
        proof = flatmap_optional(
            aggregate,
            lambda ag: make_constant_chaum_pedersen(
                message=_ciphertext_ballot_elgamal_accumulate(ballot_selections),
                constant=number_elected,
                r=ag,
                k=elgamal_public_key,
                seed=proof_seed,
                hash_header=crypto_extended_base_hash,
            ),
        )
    return CiphertextBallotContest(
        object_id=object_id,
        description_hash=description_hash,
        ballot_selections=ballot_selections,
        nonce=nonce,
        crypto_hash=crypto_hash,
        proof=proof,
    )


@dataclass(unsafe_hash=True)
class PlaintextBallot(ElectionObjectBase):
    """
    A PlaintextBallot represents a voters selections for a given ballot and ballot style
    :field object_id: A unique Ballot ID that is relevant to the external system
    """

    ballot_style: str
    """The `object_id` of the `BallotStyle` in the `Election` Manifest"""

    contests: List[PlaintextBallotContest]
    """The list of contests for this ballot"""

    def is_valid(self, expected_ballot_style_id: str) -> bool:
        """
        Check if expected ballot style is valid
        :param expected_ballot_style_id: Expected ballot style id
        :return: True if valid
        """
        if self.ballot_style != expected_ballot_style_id:
            log_warning(
                f"invalid ballot_style: for: {self.object_id} expected({expected_ballot_style_id}) actual({self.ballot_style})"
            )
            return False

        return True

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, PlaintextBallot)
            and self.ballot_style == other.ballot_style
            and _list_eq(self.contests, other.contests)
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


@dataclass(unsafe_hash=True)
class CiphertextBallot(ElectionObjectBase, CryptoHashCheckable):
    """
    A CiphertextBallot represents a voters encrypted selections for a given ballot and ballot style.

    When a ballot is in it's complete, encrypted state, the `nonce` is the master nonce
    from which all other nonces can be derived to encrypt the ballot.  Allong with the `nonce`
    fields on `Ballotcontest` and `BallotSelection`, this value is sensitive.

    Don't make this directly. Use `make_ciphertext_ballot` instead.
    :field object_id: A unique Ballot ID that is relevant to the external system
    """

    ballot_style: str
    """The `object_id` of the `BallotStyle` in the `Election` Manifest"""

    description_hash: ElementModQ
    """Hash of the election metadata"""

    previous_tracking_hash: ElementModQ
    """Previous tracking hash or seed hash"""

    contests: List[CiphertextBallotContest]
    """List of contests for this ballot"""

    tracking_hash: Optional[ElementModQ]
    """Unique ballot tracking hash for this ballot"""

    timestamp: int
    """Timestamp at which the ballot encryption is generated in tick"""

    crypto_hash: ElementModQ
    """The hash of the encrypted ballot representation"""

    nonce: Optional[ElementModQ]
    """The nonce used to encrypt this ballot. Sensitive & should be treated as a secret"""

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, CiphertextBallot)
            and self.object_id == other.object_id
            and self.ballot_style == other.ballot_style
            and self.description_hash == other.description_hash
            and self.previous_tracking_hash == other.previous_tracking_hash
            and _list_eq(self.contests, other.contests)
            and self.tracking_hash == other.tracking_hash
            and self.timestamp == other.timestamp
            and self.crypto_hash == other.crypto_hash
            and self.nonce == other.nonce
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    @staticmethod
    def nonce_seed(
        description_hash: ElementModQ, object_id: str, nonce: ElementModQ
    ) -> ElementModQ:
        """
        :return: a representation of the election and the external Id in the nonce's used
        to derive other nonce values on the ballot
        """
        return hash_elems(description_hash, object_id, nonce)

    def hashed_ballot_nonce(self) -> Optional[ElementModQ]:
        """
        :return: a hash value derived from the description hash, the object id, and the nonce value
                suitable for deriving other nonce values on the ballot
        """

        if self.nonce is None:
            log_warning(
                f"missing nonce for ballot {self.object_id} could not derive from null nonce"
            )
            return None

        return self.nonce_seed(self.description_hash, self.object_id, self.nonce)

    def get_tracker_code(self) -> Optional[str]:
        """
        Get a tracker hash as a code in friendly readable words for sharing
        :return: Tracker in words or None
        """
        if not self.tracking_hash:
            return None
        return tracker_hash_to_words(self.tracking_hash)

    def crypto_hash_with(self, seed_hash: ElementModQ) -> ElementModQ:
        """
        Given an encrypted Ballot, generates a hash, suitable for rolling up
        into a hash / tracking code for an entire ballot. Of note, this particular hash examines
        the `description_hash` and `ballot_selections`, but not the proof.
        This is deliberate, allowing for the possibility of ElectionGuard variants running on
        much more limited hardware, wherein the Disjunctive Chaum-Pedersen proofs might be computed
        later on.
        """
        if len(self.contests) == 0:
            log_warning(
                f"mismatching contests state: {self.object_id} expected(some), actual(none)"
            )
            return ZERO_MOD_Q

        contest_hashes = [contest.crypto_hash for contest in self.contests]
        return hash_elems(self.object_id, seed_hash, *contest_hashes)

    def is_valid_encryption(
        self,
        seed_hash: ElementModQ,
        elgamal_public_key: ElementModP,
        crypto_extended_base_hash: ElementModQ,
    ) -> bool:
        """
        Given an encrypted Ballot, validates the encryption state against a specific seed hash and public key
        by verifying the states of this ballot's children (BallotContest's and BallotSelection's).
        Calling this function expects that the object is in a well-formed encrypted state
        with the `contests` populated with valid encrypted ballot selections,
        and the ElementModQ `description_hash` also populated.
        Specifically, the seed hash in this context is the hash of the Election Manifest, 
        or whatever `ElementModQ` was used to populate the `description_hash` field.
        """

        if seed_hash != self.description_hash:
            log_warning(
                f"mismatching ballot hash: {self.object_id} expected({str(seed_hash)}), actual({str(self.description_hash)})"
            )
            return False

        recalculated_crypto_hash = self.crypto_hash_with(seed_hash)
        if self.crypto_hash != recalculated_crypto_hash:
            log_warning(
                f"mismatching crypto hash: {self.object_id} expected({str(recalculated_crypto_hash)}), actual({str(self.crypto_hash)})"
            )
            return False

        # Check the proofs on the ballot
        valid_proofs: List[bool] = list()

        for contest in self.contests:
            for selection in contest.ballot_selections:
                valid_proofs.append(
                    selection.is_valid_encryption(
                        selection.description_hash,
                        elgamal_public_key,
                        crypto_extended_base_hash,
                    )
                )
            valid_proofs.append(
                contest.is_valid_encryption(
                    contest.description_hash,
                    elgamal_public_key,
                    crypto_extended_base_hash,
                )
            )
        return all(valid_proofs)


class BallotBoxState(Enum):
    """
    Enumeration used when marking a ballot as cast or spoiled
    """

    CAST = 1
    """
    A ballot that has been explicitly cast
    """
    SPOILED = 2
    """
    A ballot that has been explicitly spoiled
    """
    UNKNOWN = 999
    """
    A ballot whose state is unknown to ElectionGuard and will not be included in any election results
    """


@dataclass(unsafe_hash=True)
class CiphertextAcceptedBallot(CiphertextBallot):
    """
    A `CiphertextAcceptedBallot` represents a ballot that is accepted for inclusion in election results.
    An accepted ballot is or is about to be either cast or spoiled.
    The state supports the `BallotBoxState.UNKNOWN` enumeration to indicate that this object is mutable
    and has not yet been explicitly assigned a specific state.

    Note, additionally, this ballot includes all proofs but no nonces.

    Do not make this class directly. Use `make_ciphertext_accepted_ballot` or `from_ciphertext_ballot` instead.
    """

    state: BallotBoxState

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, CiphertextAcceptedBallot)
            and super.__eq__(self, other)
            and self.state == other.state
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


def make_ciphertext_ballot(
    object_id: str,
    ballot_style: str,
    description_hash: ElementModQ,
    previous_tracking_hash: Optional[ElementModQ],
    contests: List[CiphertextBallotContest],
    nonce: Optional[ElementModQ] = None,
    timestamp: Optional[int] = None,
    tracking_hash: Optional[ElementModQ] = None,
) -> CiphertextBallot:
    """
    Makes a `CiphertextBallot`, initially in the state where it's neither been cast nor spoiled.

    :param object_id: the object_id of this specific ballot
    :param ballot_style: The `object_id` of the `BallotStyle` in the `Election` Manifest
    :param description_hash: Hash of the election metadata
    :param crypto_base_hash: Hash of the cryptographic election context
    :param contests: List of contests for this ballot
    :param timestamp: Timestamp at which the ballot encryption is generated in tick
    :param previous_tracking_hash: Previous tracking hash or seed hash
    :param nonce: optional nonce used as part of the encryption process
    """

    if len(contests) == 0:
        log_warning(f"ciphertext ballot with no contests")

    contest_hashes = [contest.crypto_hash for contest in contests]
    contest_hash = hash_elems(object_id, description_hash, *contest_hashes)

    timestamp = to_ticks(datetime.utcnow()) if timestamp is None else timestamp
    if previous_tracking_hash is None:
        previous_tracking_hash = description_hash
    if tracking_hash is None:
        tracking_hash = get_rotating_tracker_hash(
            previous_tracking_hash, timestamp, contest_hash
        )

    return CiphertextBallot(
        object_id=object_id,
        ballot_style=ballot_style,
        description_hash=description_hash,
        previous_tracking_hash=previous_tracking_hash,
        contests=contests,
        tracking_hash=tracking_hash,
        timestamp=timestamp,
        nonce=nonce,
        crypto_hash=contest_hash,
    )


def make_ciphertext_accepted_ballot(
    object_id: str,
    ballot_style: str,
    description_hash: ElementModQ,
    previous_tracking_hash: Optional[ElementModQ],
    contests: List[CiphertextBallotContest],
    tracking_hash: Optional[ElementModQ],
    timestamp: Optional[int] = None,
    state: BallotBoxState = BallotBoxState.UNKNOWN,
) -> CiphertextAcceptedBallot:
    """
    Makes a `CiphertextAcceptedBallot`, ensuring that no nonces are part of the contests.

    :param object_id: the object_id of this specific ballot
    :param ballot_style: The `object_id` of the `BallotStyle` in the `Election` Manifest
    :param description_hash: Hash of the election metadata
    :param previous_tracking_hash: Previous tracking hash or seed hash
    :param contests: List of contests for this ballot
    :param timestamp: Timestamp at which the ballot encryption is generated in tick
    :param state: ballot box state
    """

    if len(contests) == 0:
        log_warning(f"ciphertext ballot with no contests")

    contest_hashes = [contest.crypto_hash for contest in contests]
    contest_hash = hash_elems(object_id, description_hash, *contest_hashes)

    timestamp = to_ticks(datetime.utcnow()) if timestamp is None else timestamp
    if previous_tracking_hash is None:
        previous_tracking_hash = description_hash
    if tracking_hash is None:
        tracking_hash = get_rotating_tracker_hash(
            previous_tracking_hash, timestamp, contest_hash
        )

    # copy the contests and selections, removing all nonces
    new_contests: List[CiphertextBallotContest] = []
    for contest in contests:
        new_selections = [
            replace(selection, nonce=None) for selection in contest.ballot_selections
        ]
        new_contest = replace(contest, nonce=None, ballot_selections=new_selections)
        new_contests.append(new_contest)

    return CiphertextAcceptedBallot(
        object_id=object_id,
        ballot_style=ballot_style,
        description_hash=description_hash,
        previous_tracking_hash=previous_tracking_hash,
        contests=new_contests,
        tracking_hash=tracking_hash,
        timestamp=timestamp,
        crypto_hash=contest_hash,
        nonce=None,
        state=state,
    )


def from_ciphertext_ballot(
    ballot: CiphertextBallot, state: BallotBoxState = BallotBoxState.UNKNOWN
) -> CiphertextAcceptedBallot:
    """
    Convert a `CiphertextBallot` into a `CiphertextAcceptedBallot`, with all nonces removed.
    """
    return make_ciphertext_accepted_ballot(
        object_id=ballot.object_id,
        ballot_style=ballot.ballot_style,
        description_hash=ballot.description_hash,
        contests=ballot.contests,
        timestamp=ballot.timestamp,
        previous_tracking_hash=ballot.previous_tracking_hash,
        tracking_hash=ballot.tracking_hash,
        state=state,
    )



from dataclasses import dataclass

from .elgamal import ElGamalKeyPair
from .group import (
    ElementModQ,
    ElementModP,
    g_pow_p,
    mult_p,
    pow_p,
    a_plus_bc_q,
)
from .hash import hash_elems
from .logs import log_warning
from .proof import Proof, ProofUsage


@dataclass(frozen=True)
class SchnorrProof(Proof):
    """
    Representation of a Schnorr proof
    """

    public_key: ElementModP
    """k in the spec"""
    commitment: ElementModP
    """h in the spec"""
    challenge: ElementModQ
    """c in the spec"""
    response: ElementModQ
    """u in the spec"""
    usage: ProofUsage = ProofUsage.SecretValue

    def __post_init__(self) -> None:
        super().__init__()

    def is_valid(self) -> bool:
        """
        Check validity of the `proof` for proving possession of the private key corresponding
        to `public_key`.

        :return: true if the transcript is valid, false if anything is wrong
        """
        k = self.public_key
        h = self.commitment
        u = self.response
        valid_public_key = k.is_valid_residue()
        in_bounds_h = h.is_in_bounds()
        in_bounds_u = u.is_in_bounds()

        c = hash_elems(k, h)
        valid_proof = g_pow_p(u) == mult_p(h, pow_p(k, c))

        success = valid_public_key and in_bounds_h and in_bounds_u and valid_proof
        if not success:
            log_warning(
                "found an invalid Schnorr proof: %s",
                str(
                    {
                        "in_bounds_h": in_bounds_h,
                        "in_bounds_u": in_bounds_u,
                        "valid_public_key": valid_public_key,
                        "valid_proof": valid_proof,
                        "proof": self,
                    }
                ),
            )
        return success


def make_schnorr_proof(keypair: ElGamalKeyPair, r: ElementModQ) -> SchnorrProof:
    """
    Given an ElGamal keypair and a nonce, generates a proof that the prover knows the secret key without revealing it.

    :param keypair: An ElGamal keypair.
    :param r: A random element in [0,Q).
    """

    k = keypair.public_key
    h = g_pow_p(r)
    c = hash_elems(k, h)
    u = a_plus_bc_q(r, keypair.secret_key, c)

    return SchnorrProof(k, h, c, u)



from typing import TypeVar, Callable

from hypothesis.strategies import composite, SearchStrategy

from electionguard.elgamal import elgamal_keypair_from_secret
from electionguard.group import ONE_MOD_Q, TWO_MOD_Q
from electionguardtest.group import elements_mod_q_no_zero

_T = TypeVar("_T")
_DrawType = Callable[[SearchStrategy[_T]], _T]


@composite
def elgamal_keypairs(draw: _DrawType):
    """
    Generates an arbitrary ElGamal secret/public keypair.

    :param draw: Hidden argument, used by Hypothesis.
    """
    e = draw(elements_mod_q_no_zero())
    return elgamal_keypair_from_secret(e if e != ONE_MOD_Q else TWO_MOD_Q)



from dataclasses import dataclass, field
from typing import Iterable, Optional, List, Dict, Set, Tuple
from collections.abc import Container, Sized

from .ballot import (
    BallotBoxState,
    CiphertextBallotSelection,
    CiphertextAcceptedBallot,
    CiphertextSelection,
)
from .ballot_store import BallotStore
from .ballot_validator import ballot_is_valid_for_election
from .decryption_share import CiphertextDecryptionSelection
from .election import CiphertextElectionContext, InternalElectionDescription
from .election_object_base import ElectionObjectBase
from .elgamal import ElGamalCiphertext, elgamal_add
from .group import ElementModQ, ONE_MOD_P, ElementModP
from .logs import log_warning
from .scheduler import Scheduler
from .types import BALLOT_ID, CONTEST_ID, SELECTION_ID


@dataclass
class PlaintextTallySelection(ElectionObjectBase):
    """
    A plaintext Tally Selection is a decrypted selection of a contest
    """

    tally: int
    # g^tally or M in the spec
    value: ElementModP

    message: ElGamalCiphertext

    shares: List[CiphertextDecryptionSelection]


@dataclass
class CiphertextTallySelection(ElectionObjectBase, CiphertextSelection):
    """
    a CiphertextTallySelection is a homomorphic accumulation of all of the 
    CiphertextBallotSelection instances for a specific selection in an election.
    """

    description_hash: ElementModQ
    """
    The SelectionDescription hash
    """

    ciphertext: ElGamalCiphertext = field(
        default=ElGamalCiphertext(ONE_MOD_P, ONE_MOD_P)
    )
    """
    The encrypted representation of the total of all ballots for this selection
    """

    def elgamal_accumulate(
        self, elgamal_ciphertext: ElGamalCiphertext
    ) -> ElGamalCiphertext:
        """
        Homomorphically add the specified value to the message
        """
        new_value = elgamal_add(self.ciphertext, elgamal_ciphertext)
        self.ciphertext = new_value
        return self.ciphertext


@dataclass
class PlaintextTallyContest(ElectionObjectBase):
    """
    A plaintext Tally Contest is a collection of plaintext selections
    """

    selections: Dict[SELECTION_ID, PlaintextTallySelection]


@dataclass
class CiphertextTallyContest(ElectionObjectBase):
    """
    A CiphertextTallyContest is a container for associating a collection of CiphertextTallySelection
    to a specific ContestDescription
    """

    description_hash: ElementModQ
    """
    The ContestDescription hash
    """

    tally_selections: Dict[SELECTION_ID, CiphertextTallySelection]
    """
    A collection of CiphertextTallySelection mapped by SelectionDescription.object_id
    """

    def accumulate_contest(
        self,
        contest_selections: List[CiphertextBallotSelection],
        scheduler: Optional[Scheduler] = None,
    ) -> bool:
        """
        Accumulate the contest selections of an individual ballot into this tally
        """

        if len(contest_selections) == 0:
            log_warning(
                f"accumulate cannot add missing selections for contest {self.object_id}"
            )
            return False

        # Validate the input data by comparing the selection id's provided
        # to the valid selection id's for this tally contest
        selection_ids = set(
            [
                selection.object_id
                for selection in contest_selections
                if not selection.is_placeholder_selection
            ]
        )

        if any(set(self.tally_selections).difference(selection_ids)):
            log_warning(
                f"accumulate cannot add mismatched selections for contest {self.object_id}"
            )
            return False

        if scheduler is None:
            scheduler = Scheduler()

        # iterate through the tally selections and add the new value to the total
        results: List[
            Tuple[SELECTION_ID, Optional[ElGamalCiphertext]]
        ] = scheduler.schedule(
            self._accumulate_selections,
            [
                (key, selection_tally, contest_selections)
                for (key, selection_tally) in self.tally_selections.items()
            ],
        )

        for (key, ciphertext) in results:
            if ciphertext is None:
                return False
            else:
                self.tally_selections[key].ciphertext = ciphertext

        return True

    def _accumulate_selections(
        self,
        key: SELECTION_ID,
        selection_tally: CiphertextTallySelection,
        contest_selections: List[CiphertextBallotSelection],
    ) -> Tuple[SELECTION_ID, Optional[ElGamalCiphertext]]:
        use_selection = None
        for selection in contest_selections:
            if key == selection.object_id:
                use_selection = selection
                break

        # a selection on the ballot that is required was not found
        # this should never happen when using the `CiphertextTally`
        # but sanity check anyway
        if not use_selection:
            log_warning(f"add cannot accumulate for missing selection {key}")
            return key, None

        return key, selection_tally.elgamal_accumulate(use_selection.ciphertext)


@dataclass
class PlaintextTally(ElectionObjectBase):
    """
    The plaintext representation of all contests in the election
    """

    contests: Dict[CONTEST_ID, PlaintextTallyContest]

    spoiled_ballots: Dict[BALLOT_ID, Dict[CONTEST_ID, PlaintextTallyContest]]


@dataclass
class PublishedPlaintextTally(ElectionObjectBase):
    """
    The published plaintext representation of all contests in the election
    """

    contests: Dict[CONTEST_ID, PlaintextTallyContest]


def publish_plaintext_tally(tally: PlaintextTally) -> PublishedPlaintextTally:
    """Publish a plaintext tally with simpler format"""
    return PublishedPlaintextTally(tally.object_id, tally.contests)


@dataclass
class CiphertextTally(ElectionObjectBase, Container, Sized):
    """
    A `CiphertextTally` accepts cast and spoiled ballots and accumulates a tally on the cast ballots
    """

    _metadata: InternalElectionDescription
    _encryption: CiphertextElectionContext

    # A local cache of ballots id's that have already been cast
    _cast_ballot_ids: Set[BALLOT_ID] = field(init=False)

    cast: Dict[CONTEST_ID, CiphertextTallyContest] = field(init=False)
    """
    A collection of each contest and selection in an election.  
    Retains an encrypted representation of a tally for each selection
    """

    spoiled_ballots: Dict[BALLOT_ID, CiphertextAcceptedBallot] = field(
        default_factory=lambda: {}
    )
    """
    All of the ballots marked spoiled in the election
    """

    def __post_init__(self) -> None:
        object.__setattr__(self, "_cast_ballot_ids", set())
        object.__setattr__(self, "cast", self._build_tally_collection(self._metadata))

    def __len__(self) -> int:
        return len(self._cast_ballot_ids) + len(self.spoiled_ballots)

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, CiphertextAcceptedBallot):
            return False

        if (
            item.object_id in self._cast_ballot_ids
            or item.object_id in self.spoiled_ballots
        ):
            return True

        return False

    def append(self, ballot: CiphertextAcceptedBallot) -> bool:
        """
        Append a ballot to the tally and recalculate the tally.
        """
        if ballot.state == BallotBoxState.UNKNOWN:
            log_warning(f"append cannot add {ballot.object_id} with invalid state")
            return False

        if self.__contains__(ballot):
            log_warning(f"append cannot add {ballot.object_id} that is already tallied")
            return False

        if not ballot_is_valid_for_election(ballot, self._metadata, self._encryption):
            return False

        if ballot.state == BallotBoxState.CAST:
            return self._add_cast(ballot)

        if ballot.state == BallotBoxState.SPOILED:
            return self._add_spoiled(ballot)

        log_warning(f"append cannot add {ballot.object_id}")
        return False

    SELECTION_ID = str

    def batch_append(self, ballots: Iterable[CiphertextAcceptedBallot]) -> bool:
        """
        Append a collection of Ballots to the tally and recalculate
        """
        cast_ballot_selections: Dict[
            SELECTION_ID, Dict[BALLOT_ID, ElGamalCiphertext]
        ] = {}
        for ballot in ballots:
            if not self.__contains__(ballot) and ballot_is_valid_for_election(
                ballot, self._metadata, self._encryption
            ):
                if ballot.state == BallotBoxState.CAST:

                    # collect the selections so they can can be accumulated in parallel
                    for contest in ballot.contests:
                        for selection in contest.ballot_selections:
                            if selection.object_id not in cast_ballot_selections:
                                cast_ballot_selections[selection.object_id] = {}

                            cast_ballot_selections[selection.object_id][
                                ballot.object_id
                            ] = selection.ciphertext

                # just append the spoiled ballots
                elif ballot.state == BallotBoxState.SPOILED:
                    self._add_spoiled(ballot)

        # cache the cast ballot id's so they are not double counted
        if self._execute_accumulate(cast_ballot_selections):
            for ballot in ballots:
                if ballot.state == BallotBoxState.CAST:
                    self._cast_ballot_ids.add(ballot.object_id)
            return True

        return False

    def count(self) -> int:
        """
        Get a Count of the cast ballots
        """
        return len(self._cast_ballot_ids)

    @staticmethod
    def _accumulate(
        id: str, ballot_selections: Dict[BALLOT_ID, ElGamalCiphertext]
    ) -> Tuple[str, ElGamalCiphertext]:
        return (
            id,
            elgamal_add(*[ciphertext for ciphertext in ballot_selections.values()]),
        )

    def _add_cast(self, ballot: CiphertextAcceptedBallot) -> bool:
        """
        Add a cast ballot to the tally, synchronously
        """

        # iterate through the contests and elgamal add
        for contest in ballot.contests:
            # This should never happen since the ballot is validated against the election metadata
            # but it's possible the local dictionary was modified so we double check.
            if not contest.object_id in self.cast:
                log_warning(
                    f"add cast missing contest in valid set {contest.object_id}"
                )
                return False

            use_contest = self.cast[contest.object_id]
            if not use_contest.accumulate_contest(contest.ballot_selections):
                return False

            self.cast[contest.object_id] = use_contest
        self._cast_ballot_ids.add(ballot.object_id)
        return True

    def _add_spoiled(self, ballot: CiphertextAcceptedBallot) -> bool:
        """
        Add a spoiled ballot
        """

        self.spoiled_ballots[ballot.object_id] = ballot
        return True

    @staticmethod
    def _build_tally_collection(
        description: InternalElectionDescription,
    ) -> Dict[CONTEST_ID, CiphertextTallyContest]:
        """
        Build the object graph for the tally from the InternalElectionDescription
        """

        cast_collection: Dict[str, CiphertextTallyContest] = {}
        for contest in description.contests:
            # build a collection of valid selections for the contest description
            # note: we explicitly ignore the Placeholder Selections.
            contest_selections: Dict[str, CiphertextTallySelection] = {}
            for selection in contest.ballot_selections:
                contest_selections[selection.object_id] = CiphertextTallySelection(
                    selection.object_id, selection.crypto_hash()
                )

            cast_collection[contest.object_id] = CiphertextTallyContest(
                contest.object_id, contest.crypto_hash(), contest_selections
            )

        return cast_collection

    def _execute_accumulate(
        self,
        ciphertext_selections_by_selection_id: Dict[
            str, Dict[BALLOT_ID, ElGamalCiphertext]
        ],
    ) -> bool:

        result_set: List[Tuple[SELECTION_ID, ElGamalCiphertext]]
        scheduler = Scheduler()
        result_set = scheduler.schedule(
            self._accumulate,
            [
                (selection_id, selections)
                for (
                    selection_id,
                    selections,
                ) in ciphertext_selections_by_selection_id.items()
            ],
        )

        result_dict = {
            selection_id: ciphertext for (selection_id, ciphertext) in result_set
        }

        for contest_id, contest in self.cast.items():
            for selection_id, selection in contest.tally_selections.items():
                if selection_id in result_dict:
                    selection.elgamal_accumulate(result_dict[selection_id])

        return True


@dataclass
class PublishedCiphertextTally(ElectionObjectBase):
    """
    The published plaintext representation of all contests in the election
    """

    cast: Dict[CONTEST_ID, CiphertextTallyContest]


def publish_ciphertext_tally(tally: CiphertextTally) -> PublishedCiphertextTally:
    """Publish a ciphertext tally with simpler format"""
    return PublishedCiphertextTally(tally.object_id, tally.cast)


def tally_ballot(
    ballot: CiphertextAcceptedBallot, tally: CiphertextTally
) -> Optional[CiphertextTally]:
    """
    Tally a ballot that is either Cast or Spoiled
    :return: The mutated CiphertextTally or None if there is an error
    """

    if ballot.state == BallotBoxState.UNKNOWN:
        log_warning(
            f"tally ballots error tallying unknown state for ballot {ballot.object_id}"
        )
        return None

    if tally.append(ballot):
        return tally

    return None


def tally_ballots(
    store: BallotStore,
    metadata: InternalElectionDescription,
    context: CiphertextElectionContext,
) -> Optional[CiphertextTally]:
    """
    Tally all of the ballots in the ballot store.
    :return: a CiphertextTally or None if there is an error
    """
    # TODO: ISSUE #14: unique Id for the tally
    tally: CiphertextTally = CiphertextTally("election-results", metadata, context)
    if tally.batch_append(store):
        return tally
    return None



from dataclasses import dataclass
from typing import List, Optional, NamedTuple

from .auxiliary import (
    AuxiliaryKeyPair,
    AuxiliaryDecrypt,
    AuxiliaryEncrypt,
    AuxiliaryPublicKey,
)
from .data_store import DataStore
from .election_polynomial import (
    compute_polynomial_coordinate,
    ElectionPolynomial,
    generate_polynomial,
    verify_polynomial_coordinate,
)
from .elgamal import (
    ElGamalKeyPair,
    elgamal_combine_public_keys,
    elgamal_keypair_random,
)
from .group import int_to_q, rand_q, ElementModP, ElementModQ
from .rsa import rsa_keypair, rsa_decrypt, rsa_encrypt
from .schnorr import SchnorrProof, make_schnorr_proof
from .serializable import Serializable
from .types import GUARDIAN_ID
from .utils import get_optional

ElectionJointKey = ElementModP


class CeremonyDetails(NamedTuple):
    """
    Details of key ceremony
    """

    number_of_guardians: int
    quorum: int


class ElectionKeyPair(NamedTuple):
    """A tuple of election key pair, proof and polynomial"""

    key_pair: ElGamalKeyPair
    proof: SchnorrProof
    polynomial: ElectionPolynomial


class ElectionPublicKey(NamedTuple):
    """A tuple of election public key and owner information"""

    owner_id: GUARDIAN_ID
    proof: SchnorrProof
    key: ElementModP


class PublicKeySet(NamedTuple):
    """Public key set of auxiliary and election keys and owner information"""

    owner_id: GUARDIAN_ID
    sequence_order: int
    auxiliary_public_key: str
    election_public_key: ElementModP
    election_public_key_proof: SchnorrProof


class GuardianPair(NamedTuple):
    """Pair of guardians involved in sharing"""

    owner_id: GUARDIAN_ID
    designated_id: GUARDIAN_ID


class ElectionPartialKeyBackup(NamedTuple):
    """Election partial key backup used for key sharing"""

    owner_id: GUARDIAN_ID
    """
    The Id of the guardian that generated this backup
    """
    designated_id: GUARDIAN_ID
    """
    The Id of the guardian to receive this backup
    """
    designated_sequence_order: int
    """
    The sequence order of the designated guardian
    """

    encrypted_value: str
    """
    The encrypted coordinate corresponding to a secret election polynomial
    """

    coefficient_commitments: List[ElementModP]
    """
    The public keys `K_ij`generated from the election polynomial coefficients
    """

    coefficient_proofs: List[SchnorrProof]
    """
    the proofs of posession of the private keys for the election polynomial secret coefficients
    """


@dataclass
class CoefficientValidationSet(Serializable):
    """Set of validation pieces for election key coefficients"""

    owner_id: GUARDIAN_ID
    coefficient_commitments: List[ElementModP]
    coefficient_proofs: List[SchnorrProof]


class ElectionPartialKeyVerification(NamedTuple):
    """Verification of election partial key used in key sharing"""

    owner_id: GUARDIAN_ID
    designated_id: GUARDIAN_ID
    verifier_id: GUARDIAN_ID
    verified: bool


class ElectionPartialKeyChallenge(NamedTuple):
    """Challenge of election partial key used in key sharing"""

    owner_id: GUARDIAN_ID
    designated_id: GUARDIAN_ID
    designated_sequence_order: int
    """
    The sequence order of the designated guardian
    """

    value: int
    coefficient_commitments: List[ElementModP]
    coefficient_proofs: List[SchnorrProof]


def generate_elgamal_auxiliary_key_pair() -> AuxiliaryKeyPair:
    """
    Generate auxiliary key pair using elgamal
    :return: Auxiliary key pair
    """
    elgamal_key_pair = elgamal_keypair_random()
    return AuxiliaryKeyPair(
        str(elgamal_key_pair.secret_key.to_int()),
        str(elgamal_key_pair.public_key.to_int()),
    )


def generate_rsa_auxiliary_key_pair() -> AuxiliaryKeyPair:
    """
    Generate auxiliary key pair using RSA
    :return: Auxiliary key pair
    """
    rsa_key_pair = rsa_keypair()
    return AuxiliaryKeyPair(rsa_key_pair.private_key, rsa_key_pair.public_key)


def generate_election_key_pair(
    quorum: int, nonce: ElementModQ = None
) -> ElectionKeyPair:
    """
    Generate election key pair, proof, and polynomial
    :param quorum: Quorum of guardians needed to decrypt
    :return: Election key pair
    """
    polynomial = generate_polynomial(quorum, nonce)
    key_pair = ElGamalKeyPair(
        polynomial.coefficients[0], polynomial.coefficient_commitments[0]
    )
    proof = make_schnorr_proof(key_pair, rand_q())
    return ElectionKeyPair(key_pair, proof, polynomial)


def generate_election_partial_key_backup(
    owner_id: GUARDIAN_ID,
    polynomial: ElectionPolynomial,
    auxiliary_public_key: AuxiliaryPublicKey,
    encrypt: AuxiliaryEncrypt = rsa_encrypt,
) -> Optional[ElectionPartialKeyBackup]:
    """
    Generate election partial key backup for sharing
    :param owner_id: Owner of election key
    :param polynomial: The owner's Election polynomial
    :param auxiliary_public_key: The Auxiliary public key
    :param encrypt: Function to encrypt using auxiliary key
    :return: Election partial key backup
    """
    value = compute_polynomial_coordinate(
        auxiliary_public_key.sequence_order, polynomial
    )
    encrypted_value = encrypt(str(value.to_int()), auxiliary_public_key.key)
    if encrypted_value is None:
        return None
    return ElectionPartialKeyBackup(
        owner_id,
        auxiliary_public_key.owner_id,
        auxiliary_public_key.sequence_order,
        encrypted_value,
        polynomial.coefficient_commitments,
        polynomial.coefficient_proofs,
    )


def get_coefficient_validation_set(
    owner_id: GUARDIAN_ID, polynomial: ElectionPolynomial,
) -> CoefficientValidationSet:
    """Get coefficient validation set from polynomial"""
    return CoefficientValidationSet(
        owner_id, polynomial.coefficient_commitments, polynomial.coefficient_proofs,
    )


def get_coefficient_validation_set_from_backup(
    backup: ElectionPartialKeyBackup,
) -> CoefficientValidationSet:
    """Get coefficient validation set from a election partial key backup"""
    return CoefficientValidationSet(
        backup.owner_id, backup.coefficient_commitments, backup.coefficient_proofs,
    )


def verify_election_partial_key_backup(
    verifier_id: GUARDIAN_ID,
    backup: ElectionPartialKeyBackup,
    auxiliary_key_pair: AuxiliaryKeyPair,
    decrypt: AuxiliaryDecrypt = rsa_decrypt,
) -> ElectionPartialKeyVerification:
    """
    Verify election partial key backup contain point on owners polynomial
    :param verifier_id: Verifier of the partial key backup
    :param backup: Election partial key backup
    :param auxiliary_key_pair: Auxiliary key pair
    :param decrypt: Decryption function using auxiliary key
    """

    decrypted_value = decrypt(backup.encrypted_value, auxiliary_key_pair.secret_key)
    if decrypted_value is None:
        return ElectionPartialKeyVerification(
            backup.owner_id, backup.designated_id, verifier_id, False
        )
    value = get_optional(int_to_q(int(decrypted_value)))
    return ElectionPartialKeyVerification(
        backup.owner_id,
        backup.designated_id,
        verifier_id,
        verify_polynomial_coordinate(
            value, backup.designated_sequence_order, backup.coefficient_commitments
        ),
    )


def generate_election_partial_key_challenge(
    backup: ElectionPartialKeyBackup, polynomial: ElectionPolynomial,
) -> ElectionPartialKeyChallenge:
    """
    Generate challenge to a previous verification of a partial key backup
    :param backup: Election partial key backup in question
    :param poynomial: Polynomial to regenerate point
    :return: Election partial key verification
    """
    return ElectionPartialKeyChallenge(
        backup.owner_id,
        backup.designated_id,
        backup.designated_sequence_order,
        compute_polynomial_coordinate(
            backup.designated_sequence_order, polynomial
        ).to_int(),
        backup.coefficient_commitments,
        backup.coefficient_proofs,
    )


def verify_election_partial_key_challenge(
    verifier_id: GUARDIAN_ID, challenge: ElectionPartialKeyChallenge
) -> ElectionPartialKeyVerification:
    """
    Verify a challenge to a previous verification of a partial key backup
    :param verifier_id: Verifier of the challenge
    :param challenge: Election partial key challenge
    :return: Election partial key verification
    """
    return ElectionPartialKeyVerification(
        challenge.owner_id,
        challenge.designated_id,
        verifier_id,
        verify_polynomial_coordinate(
            get_optional(int_to_q(challenge.value)),
            challenge.designated_sequence_order,
            challenge.coefficient_commitments,
        ),
    )


def combine_election_public_keys(
    election_public_keys: DataStore[GUARDIAN_ID, ElectionPublicKey]
) -> ElectionJointKey:
    """
    Creates a joint election key from the public keys of all guardians
    :return: Joint key for election
    """
    public_keys = map(lambda public_key: public_key.key, election_public_keys.values())

    return elgamal_combine_public_keys(public_keys)



from dataclasses import dataclass
from datetime import datetime
from os import path
from typing import cast, TypeVar, Generic

from jsons import (
    dumps,
    NoneType,
    loads,
    JsonsError,
    set_deserializer,
    set_serializer,
    set_validator,
    suppress_warnings,
    default_nonetype_deserializer,
)

T = TypeVar("T")

JSON_FILE_EXTENSION: str = ".json"
WRITE: str = "w"
READ: str = "r"
JSON_PARSE_ERROR = '{"error": "Object could not be parsed due to json issue"}'
# TODO Issue #??: Jsons library incorrectly dumps class method
FROM_JSON_FILE = '"from_json_file": {}, '


@dataclass
class Serializable(Generic[T]):
    """
    Serializable class with methods to convert to json
    """

    def to_json(self, strip_privates: bool = True) -> str:
        """
        Serialize to json
        :param strip_privates: strip private variables
        :return: the json representation of this object
        """
        set_serializers()
        suppress_warnings()
        try:
            return cast(
                str, dumps(self, strip_privates=strip_privates, strip_nulls=True)
            ).replace(FROM_JSON_FILE, "")
        except JsonsError:
            return JSON_PARSE_ERROR

    def to_json_file(
        self, file_name: str, file_path: str = "", strip_privates: bool = True
    ) -> None:
        """
        Serialize an object to a json file
        :param file_name: File name
        :param file_path: File path
        :param strip_privates: Strip private variables
        """
        write_json_file(self.to_json(strip_privates), file_name, file_path)

    @classmethod
    def from_json_file(cls, file_name: str, file_path: str = "") -> T:
        """
        Deserialize the provided file into the specified instance
        """
        json_file_path: str = path.join(file_path, file_name + JSON_FILE_EXTENSION)
        with open(json_file_path, READ) as json_file:
            data = json_file.read()
            target = cls.from_json(data)
        return target

    @classmethod
    def from_json(cls, data: str) -> T:
        """
        Deserialize the provided data string into the specified instance
        """
        set_deserializers()
        return cast(T, loads(data, cls))


def write_json_file(json_data: str, file_name: str, file_path: str = "") -> None:
    """
    Write json data string to json file
    """
    json_file_path: str = path.join(file_path, file_name + JSON_FILE_EXTENSION)
    with open(json_file_path, WRITE) as json_file:
        json_file.write(json_data)


def set_serializers() -> None:
    """Set serializers for jsons to use to cast specific classes"""

    # Local import to minimize jsons usage across files
    from .group import ElementModP, ElementModQ

    set_serializer(lambda p, **_: str(p), ElementModP)
    set_serializer(lambda q, **_: str(q), ElementModQ)
    set_serializer(lambda dt, **_: dt.isoformat(), datetime)


def set_deserializers() -> None:
    """Set deserializers and validators for json to use to cast specific classes"""

    # Local import to minimize jsons usage across files
    from .group import ElementModP, ElementModQ, int_to_p_unchecked, int_to_q_unchecked

    set_deserializer(
        lambda p_as_int, cls, **_: int_to_p_unchecked(p_as_int), ElementModP
    )
    set_validator(lambda p: p.is_in_bounds(), ElementModP)

    set_deserializer(
        lambda q_as_int, cls, **_: int_to_q_unchecked(q_as_int), ElementModQ
    )
    set_validator(lambda q: q.is_in_bounds(), ElementModQ)

    set_deserializer(
        lambda none, cls, **_: None
        if none == "None"
        else default_nonetype_deserializer(none),
        NoneType,
    )

    set_deserializer(lambda dt, cls, **_: datetime.fromisoformat(dt), datetime)
